@@ -4,8 +4,7 @@
 // License, v. 2.0. If a copy of the MPL was not distributed with this file,
 // You can obtain one at https://mozilla.org/MPL/2.0/.
 
-//go:generate -command counterfeiter go run github.com/maxbrunsfeld/counterfeiter/v6
-//go:generate counterfeiter -o mocks/service.go --fake-name Service . Service
+//go:generate go tool counterfeiter -o mocks/service.go --fake-name Service . Service
 
 package connections
 
@@ -1341,7 +1340,7 @@ func (c *deviceConnectionTracker) accountAddedConnection(conn protocol.Connectio
 	// how many total connections they want
 	d := conn.DeviceID()
 	c.connections[d] = append(c.connections[d], conn)
-	c.wantConnections[d] = int(h.NumConnections)
+	c.wantConnections[d] = h.NumConnections
 	l.Debugf("Added connection for %s (now %d), they want %d connections", d.Short(), len(c.connections[d]), h.NumConnections)
 
 	// Update active connections metric
